/*
 * based on: snul.vala (part of Snul)
 *
 * Snul: is a simple library that implements
 *       a tcp socket client in vala
 *
 * Copyright (C) 2008  Andrea Del Signore
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 *
 * Author:
 *     Andrea Del Signore <sejerpz@tin.it>
 */

#include "snul-gnio.h"
#include <gio/gio.h>




struct _SnulTcpClientPrivate {
	char* _address;
	char* _first_address;
	char* _port;
	gushort _port_number;
	GResolver* _resolver;
	GInetAddress* _inet_address;
	GSocketConnection* _client;
	GSource* _source;
	GDataInputStream* _input;
	GCancellable* _cancellable;
	guint _timeout_id;
};

#define SNUL_TCP_CLIENT_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SNUL_TYPE_TCP_CLIENT, SnulTcpClientPrivate))
enum  {
	SNUL_TCP_CLIENT_DUMMY_PROPERTY,
	SNUL_TCP_CLIENT_ADDRESS,
	SNUL_TCP_CLIENT_FIRST_ADDRESS,
	SNUL_TCP_CLIENT_PORT
};
static void _snul_tcp_client_on_client_connected_gasync_ready_callback (GObject* source_object, GAsyncResult* res, gpointer self);
static void snul_tcp_client_on_address_resolved (GObject* sender, GAsyncResult* result, SnulTcpClient* self);
static void snul_tcp_client_on_client_connected (GObject* sender, GAsyncResult* result, SnulTcpClient* self);
static gboolean _snul_tcp_client_on_operation_timed_out_gsource_func (gpointer self);
static void _snul_tcp_client_on_address_resolved_gasync_ready_callback (GObject* source_object, GAsyncResult* res, gpointer self);
static gboolean snul_tcp_client_on_operation_timed_out (SnulTcpClient* self);
static void snul_tcp_client_cleanup (SnulTcpClient* self);
static void snul_tcp_client_real_on_connected (SnulTcpClient* self, GSocketConnection* client);
static void snul_tcp_client_real_on_data_received (SnulTcpClient* self, const char* buffer, gsize len);
static void snul_tcp_client_real_on_disconnected (SnulTcpClient* self);
static void snul_tcp_client_real_on_error (SnulTcpClient* self, GError* error);
static gboolean snul_tcp_client_socket_callback (SnulTcpClient* self, GIOCondition condition, gint fd);
static gpointer snul_tcp_client_parent_class = NULL;
static void snul_tcp_client_dispose (GObject * obj);

static const gint SNUL_TCP_CLIENT_TimeoutValue = 10;

static void g_cclosure_user_marshal_VOID__STRING_ULONG (GClosure * closure, GValue * return_value, guint n_param_values, const GValue * param_values, gpointer invocation_hint, gpointer marshal_data);

GQuark snul_tcp_socket_error_quark (void) {
	return g_quark_from_static_string ("snul_tcp_socket_error-quark");
}


static void _snul_tcp_client_on_client_connected_gasync_ready_callback (GObject* source_object, GAsyncResult* res, gpointer self) {
	snul_tcp_client_on_client_connected (source_object, res, self);
}


static void snul_tcp_client_on_address_resolved (GObject* sender, GAsyncResult* result, SnulTcpClient* self) {
	GError * inner_error;
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	g_return_if_fail (G_IS_OBJECT (sender));
	g_return_if_fail (G_IS_ASYNC_RESULT (result));
	inner_error = NULL;
	{
		GInetAddress* _tmp1;
		GInetAddress* _tmp0;
		GInetSocketAddress* socket_address;
		GSocketConnection* _tmp2;
		_tmp1 = NULL;
		_tmp0 = NULL;
		self->priv->_inet_address = (_tmp1 = (_tmp0 = g_resolver_resolve_finish ((G_RESOLVER (sender)), result, &inner_error), (_tmp0 == NULL ? NULL : g_object_ref (_tmp0))), (self->priv->_inet_address == NULL ? NULL : (self->priv->_inet_address = (g_object_unref (self->priv->_inet_address), NULL))), _tmp1);
		if (inner_error != NULL) {
			goto __catch10_g_error;
		}
		socket_address = g_object_ref_sink (g_inet_socket_address_new (self->priv->_inet_address, ((gushort) (atoi (self->priv->_port)))));
		_tmp2 = NULL;
		self->priv->_client = (_tmp2 = g_socket_connection_new (G_SOCKET_ADDRESS (socket_address)), (self->priv->_client == NULL ? NULL : (self->priv->_client = (g_object_unref (self->priv->_client), NULL))), _tmp2);
		g_socket_connection_connect_async (self->priv->_client, NULL, _snul_tcp_client_on_client_connected_gasync_ready_callback, self);
		(socket_address == NULL ? NULL : (socket_address = (g_object_unref (socket_address), NULL)));
	}
	goto __finally10;
	__catch10_g_error:
	{
		GError * err;
		err = inner_error;
		inner_error = NULL;
		{
			g_signal_emit_by_name (G_OBJECT (self), "error", err);
			(err == NULL ? NULL : (err = (g_error_free (err), NULL)));
		}
	}
	__finally10:
	;
}


static void snul_tcp_client_on_client_connected (GObject* sender, GAsyncResult* result, SnulTcpClient* self) {
	GError * inner_error;
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	g_return_if_fail (G_IS_OBJECT (sender));
	g_return_if_fail (G_IS_ASYNC_RESULT (result));
	inner_error = NULL;
	{
		GSocketConnection* _tmp1;
		GSocketConnection* _tmp0;
		GSocket* _tmp2;
		GDataInputStream* _tmp3;
		GIOCondition conditions;
		GSource* _tmp6;
		GSource* _tmp5;
		GSocket* _tmp4;
		GCancellable* _tmp7;
		g_socket_connection_connect_finish ((G_SOCKET_CONNECTION (sender)), result, &inner_error);
		if (inner_error != NULL) {
			goto __catch11_g_error;
		}
		/* create a source for
		 monitoring incoming data*/
		_tmp1 = NULL;
		_tmp0 = NULL;
		self->priv->_client = (_tmp1 = (_tmp0 = G_SOCKET_CONNECTION (sender), (_tmp0 == NULL ? NULL : g_object_ref (_tmp0))), (self->priv->_client == NULL ? NULL : (self->priv->_client = (g_object_unref (self->priv->_client), NULL))), _tmp1);
		_tmp2 = NULL;
		g_socket_set_blocking ((g_object_get (G_OBJECT (g_socket_connection_get_input_stream (self->priv->_client)), "socket", &_tmp2, NULL), _tmp2), FALSE);
		_tmp3 = NULL;
		self->priv->_input = (_tmp3 = g_data_input_stream_new (G_INPUT_STREAM (g_socket_connection_get_input_stream (self->priv->_client))), (self->priv->_input == NULL ? NULL : (self->priv->_input = (g_object_unref (self->priv->_input), NULL))), _tmp3);
		conditions = G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL;
		_tmp6 = NULL;
		_tmp5 = NULL;
		_tmp4 = NULL;
		self->priv->_source = (_tmp6 = (_tmp5 = g_socket_create_source ((g_object_get (G_OBJECT (g_socket_connection_get_input_stream (self->priv->_client)), "socket", &_tmp4, NULL), _tmp4), conditions, NULL), (_tmp5 == NULL ? NULL : g_source_ref (_tmp5))), (self->priv->_source == NULL ? NULL : (self->priv->_source = (g_source_unref (self->priv->_source), NULL))), _tmp6);
		g_source_set_callback (self->priv->_source, ((void*) (snul_tcp_client_socket_callback)), self, NULL);
		g_source_attach (self->priv->_source, NULL);
		g_signal_emit_by_name (G_OBJECT (self), "connected", self->priv->_client);
		_tmp7 = NULL;
		self->priv->_cancellable = (_tmp7 = NULL, (self->priv->_cancellable == NULL ? NULL : (self->priv->_cancellable = (g_object_unref (self->priv->_cancellable), NULL))), _tmp7);
		if (self->priv->_timeout_id != 0) {
			g_source_remove (self->priv->_timeout_id);
			self->priv->_timeout_id = ((guint) (0));
		}
	}
	goto __finally11;
	__catch11_g_error:
	{
		GError * sock;
		sock = inner_error;
		inner_error = NULL;
		{
			snul_tcp_client_on_error (self, sock);
			(sock == NULL ? NULL : (sock = (g_error_free (sock), NULL)));
		}
	}
	__finally11:
	;
}


static gboolean _snul_tcp_client_on_operation_timed_out_gsource_func (gpointer self) {
	return snul_tcp_client_on_operation_timed_out (self);
}


static void _snul_tcp_client_on_address_resolved_gasync_ready_callback (GObject* source_object, GAsyncResult* res, gpointer self) {
	snul_tcp_client_on_address_resolved (source_object, res, self);
}


void snul_tcp_client_connect (SnulTcpClient* self, const char* address, const char* port, GError** error) {
	GError * inner_error;
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	g_return_if_fail (address != NULL);
	g_return_if_fail (port != NULL);
	inner_error = NULL;
	{
		char* _tmp1;
		const char* _tmp0;
		char* _tmp3;
		const char* _tmp2;
		GCancellable* _tmp4;
		GResolver* _tmp5;
		if (self->priv->_client != NULL) {
			inner_error = g_error_new (SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_SOCKET_ALREADY_OPEN, "Socket already opened");
			if (inner_error != NULL) {
				goto __catch12_g_error;
			}
		}
		_tmp1 = NULL;
		_tmp0 = NULL;
		self->priv->_address = (_tmp1 = (_tmp0 = address, (_tmp0 == NULL ? NULL : g_strdup (_tmp0))), (self->priv->_address = (g_free (self->priv->_address), NULL)), _tmp1);
		_tmp3 = NULL;
		_tmp2 = NULL;
		self->priv->_port = (_tmp3 = (_tmp2 = port, (_tmp2 == NULL ? NULL : g_strdup (_tmp2))), (self->priv->_port = (g_free (self->priv->_port), NULL)), _tmp3);
		_tmp4 = NULL;
		self->priv->_cancellable = (_tmp4 = g_cancellable_new (), (self->priv->_cancellable == NULL ? NULL : (self->priv->_cancellable = (g_object_unref (self->priv->_cancellable), NULL))), _tmp4);
		self->priv->_timeout_id = g_timeout_add_seconds (((guint) (SNUL_TCP_CLIENT_TimeoutValue)), _snul_tcp_client_on_operation_timed_out_gsource_func, self);
		_tmp5 = NULL;
		self->priv->_resolver = (_tmp5 = g_resolver_new (), (self->priv->_resolver == NULL ? NULL : (self->priv->_resolver = (g_object_unref (self->priv->_resolver), NULL))), _tmp5);
		g_resolver_resolve_async (self->priv->_resolver, self->priv->_address, self->priv->_cancellable, _snul_tcp_client_on_address_resolved_gasync_ready_callback, self);
	}
	goto __finally12;
	__catch12_g_error:
	{
		GError * err;
		err = inner_error;
		inner_error = NULL;
		{
			GError* _tmp6;
			snul_tcp_client_on_error (self, err);
			_tmp6 = NULL;
			inner_error = (_tmp6 = err, (_tmp6 == NULL || g_error_copy == NULL ? ((gpointer) (_tmp6)) : g_error_copy (((gpointer) (_tmp6)))));
			if (inner_error != NULL) {
				g_propagate_error (error, inner_error);
				(err == NULL ? NULL : (err = (g_error_free (err), NULL)));
				return;
			}
			(err == NULL ? NULL : (err = (g_error_free (err), NULL)));
		}
	}
	__finally12:
	;
}


static gboolean snul_tcp_client_on_operation_timed_out (SnulTcpClient* self) {
	g_return_val_if_fail (SNUL_IS_TCP_CLIENT (self), FALSE);
	if (self->priv->_cancellable != NULL && !g_cancellable_is_cancelled (self->priv->_cancellable)) {
		GError* _tmp0;
		g_cancellable_cancel (self->priv->_cancellable);
		_tmp0 = NULL;
		snul_tcp_client_on_error (self, (_tmp0 = g_error_new (SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_OPERATION_TIMEOUT, "Operation timeout")));
		(_tmp0 == NULL ? NULL : (_tmp0 = (g_error_free (_tmp0), NULL)));
	}
	self->priv->_timeout_id = ((guint) (0));
	return FALSE;
}


gsize snul_tcp_client_send (SnulTcpClient* self, const char* buffer) {
	GError * inner_error;
	gsize bytes_written;
	g_return_val_if_fail (SNUL_IS_TCP_CLIENT (self), 0UL);
	g_return_val_if_fail (buffer != NULL, 0UL);
	inner_error = NULL;
	bytes_written = ((gsize) (0));
	{
		g_output_stream_write (G_OUTPUT_STREAM (g_socket_connection_get_output_stream (self->priv->_client)), buffer, ((gulong) (g_utf8_strlen (buffer, -1))), NULL, &inner_error);
		if (inner_error != NULL) {
			goto __catch13_g_error;
		}
	}
	goto __finally13;
	__catch13_g_error:
	{
		GError * err;
		err = inner_error;
		inner_error = NULL;
		{
			g_warning ("snul-gnio.vala:154: error writing: %s", err->message);
			snul_tcp_client_on_error (self, err);
			(err == NULL ? NULL : (err = (g_error_free (err), NULL)));
		}
	}
	__finally13:
	;
	return bytes_written;
}


void snul_tcp_client_disconnect (SnulTcpClient* self) {
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	snul_tcp_client_cleanup (self);
	snul_tcp_client_on_disconnected (self);
}


static void snul_tcp_client_cleanup (SnulTcpClient* self) {
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	g_source_remove_by_user_data (self);
	if (self->priv->_source != NULL) {
		GSource* _tmp0;
		g_source_destroy (self->priv->_source);
		_tmp0 = NULL;
		self->priv->_source = (_tmp0 = NULL, (self->priv->_source == NULL ? NULL : (self->priv->_source = (g_source_unref (self->priv->_source), NULL))), _tmp0);
	}
	if (self->priv->_client != NULL) {
		GDataInputStream* _tmp1;
		GSocketConnection* _tmp2;
		_tmp1 = NULL;
		self->priv->_input = (_tmp1 = NULL, (self->priv->_input == NULL ? NULL : (self->priv->_input = (g_object_unref (self->priv->_input), NULL))), _tmp1);
		g_socket_connection_close (self->priv->_client);
		_tmp2 = NULL;
		self->priv->_client = (_tmp2 = NULL, (self->priv->_client == NULL ? NULL : (self->priv->_client = (g_object_unref (self->priv->_client), NULL))), _tmp2);
	}
}


static void snul_tcp_client_real_on_connected (SnulTcpClient* self, GSocketConnection* client) {
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	g_return_if_fail (G_IS_SOCKET_CONNECTION (client));
	g_signal_emit_by_name (G_OBJECT (self), "connected", client);
}


void snul_tcp_client_on_connected (SnulTcpClient* self, GSocketConnection* client) {
	SNUL_TCP_CLIENT_GET_CLASS (self)->on_connected (self, client);
}


static void snul_tcp_client_real_on_data_received (SnulTcpClient* self, const char* buffer, gsize len) {
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	g_return_if_fail (buffer != NULL);
	g_signal_emit_by_name (G_OBJECT (self), "data-received", buffer, len);
}


void snul_tcp_client_on_data_received (SnulTcpClient* self, const char* buffer, gsize len) {
	SNUL_TCP_CLIENT_GET_CLASS (self)->on_data_received (self, buffer, len);
}


static void snul_tcp_client_real_on_disconnected (SnulTcpClient* self) {
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	g_signal_emit_by_name (G_OBJECT (self), "disconnected");
}


void snul_tcp_client_on_disconnected (SnulTcpClient* self) {
	SNUL_TCP_CLIENT_GET_CLASS (self)->on_disconnected (self);
}


static void snul_tcp_client_real_on_error (SnulTcpClient* self, GError* error) {
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	g_signal_emit_by_name (G_OBJECT (self), "error", error);
}


void snul_tcp_client_on_error (SnulTcpClient* self, GError* error) {
	SNUL_TCP_CLIENT_GET_CLASS (self)->on_error (self, error);
}


static gboolean snul_tcp_client_socket_callback (SnulTcpClient* self, GIOCondition condition, gint fd) {
	GError * inner_error;
	gboolean res;
	GIOCondition _tmp5;
	g_return_val_if_fail (SNUL_IS_TCP_CLIENT (self), FALSE);
	inner_error = NULL;
	res = TRUE;
	_tmp5 = condition;
	if (_tmp5 == G_IO_IN)
	do {
		gsize length;
		GString* message;
		length = ((gsize) (-1));
		message = g_string_new ("");
		while (length != 0) {
			{
				const char* _tmp0;
				char* buffer;
				_tmp0 = NULL;
				buffer = (_tmp0 = g_data_input_stream_read_line (self->priv->_input, &length, NULL, &inner_error), (_tmp0 == NULL ? NULL : g_strdup (_tmp0)));
				if (inner_error != NULL) {
					if (inner_error->domain == G_IO_ERROR) {
						goto __catch14_g_io_error;
					}
					g_critical ("file %s: line %d: uncaught error: %s", __FILE__, __LINE__, inner_error->message);
					g_clear_error (&inner_error);
				}
				if (buffer != NULL) {
					g_string_append_printf (message, "%s\n", buffer);
				}
				buffer = (g_free (buffer), NULL);
			}
			goto __finally14;
			__catch14_g_io_error:
			{
				GError * err;
				err = inner_error;
				inner_error = NULL;
				{
					if (err->code != 27) {
						/*IOError.WOULD_BLOCK*/
						snul_tcp_client_on_error (self, err);
					}
					length = ((gsize) (0));
					(err == NULL ? NULL : (err = (g_error_free (err), NULL)));
				}
			}
			__finally14:
			;
		}
		if (message->len > 0) {
			snul_tcp_client_on_data_received (self, message->str, ((gsize) (message->len)));
		} else {
			GError* _tmp1;
			_tmp1 = NULL;
			snul_tcp_client_on_error (self, (_tmp1 = g_error_new (SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_SOCKET_BROKEN, "zero length data, socket broken? Closing connection.")));
			(_tmp1 == NULL ? NULL : (_tmp1 = (g_error_free (_tmp1), NULL)));
			snul_tcp_client_disconnect (self);
		}
		(message == NULL ? NULL : (message = (g_string_free (message, TRUE), NULL)));
		break;
	} while (0); else if (_tmp5 == G_IO_HUP)
	do {
		GError* _tmp2;
		_tmp2 = NULL;
		snul_tcp_client_on_error (self, (_tmp2 = g_error_new (SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_SOCKET_HUP, "Socket HUP")));
		(_tmp2 == NULL ? NULL : (_tmp2 = (g_error_free (_tmp2), NULL)));
		snul_tcp_client_disconnect (self);
		res = FALSE;
		break;
	} while (0); else if (_tmp5 == G_IO_ERR)
	do {
		GError* _tmp3;
		_tmp3 = NULL;
		snul_tcp_client_on_error (self, (_tmp3 = g_error_new (SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_IOCHANNEL_ERROR, "IOChannel error (ERR)")));
		(_tmp3 == NULL ? NULL : (_tmp3 = (g_error_free (_tmp3), NULL)));
		snul_tcp_client_disconnect (self);
		break;
	} while (0); else if (_tmp5 == G_IO_NVAL)
	do {
		GError* _tmp4;
		_tmp4 = NULL;
		snul_tcp_client_on_error (self, (_tmp4 = g_error_new (SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_IOCHANNEL_NVAL, "IOChannel invalid request (NAVL)")));
		(_tmp4 == NULL ? NULL : (_tmp4 = (g_error_free (_tmp4), NULL)));
		snul_tcp_client_disconnect (self);
		break;
	} while (0);
	return res;
}


SnulTcpClient* snul_tcp_client_new (void) {
	SnulTcpClient * self;
	self = g_object_newv (SNUL_TYPE_TCP_CLIENT, 0, NULL);
	return self;
}


const char* snul_tcp_client_get_address (SnulTcpClient* self) {
	g_return_val_if_fail (SNUL_IS_TCP_CLIENT (self), NULL);
	return self->priv->_address;
}


const char* snul_tcp_client_get_first_address (SnulTcpClient* self) {
	g_return_val_if_fail (SNUL_IS_TCP_CLIENT (self), NULL);
	return snul_tcp_client_get_address (self);
}


void snul_tcp_client_set_first_address (SnulTcpClient* self, const char* value) {
	char* _tmp2;
	const char* _tmp1;
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	_tmp2 = NULL;
	_tmp1 = NULL;
	self->priv->_first_address = (_tmp2 = (_tmp1 = value, (_tmp1 == NULL ? NULL : g_strdup (_tmp1))), (self->priv->_first_address = (g_free (self->priv->_first_address), NULL)), _tmp2);
	g_object_notify (((GObject *) (self)), "first-address");
}


const char* snul_tcp_client_get_port (SnulTcpClient* self) {
	g_return_val_if_fail (SNUL_IS_TCP_CLIENT (self), NULL);
	return self->priv->_port;
}


static void snul_tcp_client_get_property (GObject * object, guint property_id, GValue * value, GParamSpec * pspec) {
	SnulTcpClient * self;
	self = SNUL_TCP_CLIENT (object);
	switch (property_id) {
		case SNUL_TCP_CLIENT_ADDRESS:
		g_value_set_string (value, snul_tcp_client_get_address (self));
		break;
		case SNUL_TCP_CLIENT_FIRST_ADDRESS:
		g_value_set_string (value, snul_tcp_client_get_first_address (self));
		break;
		case SNUL_TCP_CLIENT_PORT:
		g_value_set_string (value, snul_tcp_client_get_port (self));
		break;
		default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}


static void snul_tcp_client_set_property (GObject * object, guint property_id, const GValue * value, GParamSpec * pspec) {
	SnulTcpClient * self;
	self = SNUL_TCP_CLIENT (object);
	switch (property_id) {
		case SNUL_TCP_CLIENT_FIRST_ADDRESS:
		snul_tcp_client_set_first_address (self, g_value_get_string (value));
		break;
		default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}


static void snul_tcp_client_class_init (SnulTcpClientClass * klass) {
	snul_tcp_client_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (SnulTcpClientPrivate));
	G_OBJECT_CLASS (klass)->get_property = snul_tcp_client_get_property;
	G_OBJECT_CLASS (klass)->set_property = snul_tcp_client_set_property;
	G_OBJECT_CLASS (klass)->dispose = snul_tcp_client_dispose;
	SNUL_TCP_CLIENT_CLASS (klass)->on_connected = snul_tcp_client_real_on_connected;
	SNUL_TCP_CLIENT_CLASS (klass)->on_data_received = snul_tcp_client_real_on_data_received;
	SNUL_TCP_CLIENT_CLASS (klass)->on_disconnected = snul_tcp_client_real_on_disconnected;
	SNUL_TCP_CLIENT_CLASS (klass)->on_error = snul_tcp_client_real_on_error;
	g_object_class_install_property (G_OBJECT_CLASS (klass), SNUL_TCP_CLIENT_ADDRESS, g_param_spec_string ("address", "address", "address", NULL, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), SNUL_TCP_CLIENT_FIRST_ADDRESS, g_param_spec_string ("first-address", "first-address", "first-address", NULL, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), SNUL_TCP_CLIENT_PORT, g_param_spec_string ("port", "port", "port", NULL, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE));
	g_signal_new ("connected", SNUL_TYPE_TCP_CLIENT, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__OBJECT, G_TYPE_NONE, 1, G_TYPE_SOCKET_CONNECTION);
	g_signal_new ("data_received", SNUL_TYPE_TCP_CLIENT, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_user_marshal_VOID__STRING_ULONG, G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_ULONG);
	g_signal_new ("disconnected", SNUL_TYPE_TCP_CLIENT, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);
	g_signal_new ("error", SNUL_TYPE_TCP_CLIENT, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__POINTER, G_TYPE_NONE, 1, G_TYPE_POINTER);
}


static void snul_tcp_client_instance_init (SnulTcpClient * self) {
	self->priv = SNUL_TCP_CLIENT_GET_PRIVATE (self);
	self->priv->_address = NULL;
	self->priv->_first_address = NULL;
	self->priv->_port = NULL;
	self->priv->_port_number = ((gushort) (0));
	self->priv->_resolver = NULL;
	self->priv->_inet_address = NULL;
	self->priv->_client = NULL;
	self->priv->_source = NULL;
	self->priv->_input = NULL;
	self->priv->_cancellable = NULL;
	self->priv->_timeout_id = ((guint) (0));
}


static void snul_tcp_client_dispose (GObject * obj) {
	SnulTcpClient * self;
	self = SNUL_TCP_CLIENT (obj);
	{
		snul_tcp_client_cleanup (self);
	}
	self->priv->_address = (g_free (self->priv->_address), NULL);
	self->priv->_first_address = (g_free (self->priv->_first_address), NULL);
	self->priv->_port = (g_free (self->priv->_port), NULL);
	(self->priv->_resolver == NULL ? NULL : (self->priv->_resolver = (g_object_unref (self->priv->_resolver), NULL)));
	(self->priv->_inet_address == NULL ? NULL : (self->priv->_inet_address = (g_object_unref (self->priv->_inet_address), NULL)));
	(self->priv->_client == NULL ? NULL : (self->priv->_client = (g_object_unref (self->priv->_client), NULL)));
	(self->priv->_source == NULL ? NULL : (self->priv->_source = (g_source_unref (self->priv->_source), NULL)));
	(self->priv->_input == NULL ? NULL : (self->priv->_input = (g_object_unref (self->priv->_input), NULL)));
	(self->priv->_cancellable == NULL ? NULL : (self->priv->_cancellable = (g_object_unref (self->priv->_cancellable), NULL)));
	G_OBJECT_CLASS (snul_tcp_client_parent_class)->dispose (obj);
}


GType snul_tcp_client_get_type (void) {
	static volatile gsize snul_tcp_client_type_id = 0;
	if (g_once_init_enter (&snul_tcp_client_type_id)) {
		GType snul_tcp_client_type_id_temp;
		static const GTypeInfo g_define_type_info = { sizeof (SnulTcpClientClass), (GBaseInitFunc) NULL, (GBaseFinalizeFunc) NULL, (GClassInitFunc) snul_tcp_client_class_init, (GClassFinalizeFunc) NULL, NULL, sizeof (SnulTcpClient), 0, (GInstanceInitFunc) snul_tcp_client_instance_init };
		snul_tcp_client_type_id_temp = g_type_register_static (G_TYPE_OBJECT, "SnulTcpClient", &g_define_type_info, 0);
		g_once_init_leave (&snul_tcp_client_type_id, snul_tcp_client_type_id_temp);
	}
	return snul_tcp_client_type_id;
}



static void g_cclosure_user_marshal_VOID__STRING_ULONG (GClosure * closure, GValue * return_value, guint n_param_values, const GValue * param_values, gpointer invocation_hint, gpointer marshal_data) {
	typedef void (*GMarshalFunc_VOID__STRING_ULONG) (gpointer data1, const char* arg_1, gsize arg_2, gpointer data2);
	register GMarshalFunc_VOID__STRING_ULONG callback;
	register GCClosure * cc;
	register gpointer data1, data2;
	cc = ((GCClosure *) (closure));
	g_return_if_fail (n_param_values == 3);
	if (G_CCLOSURE_SWAP_DATA (closure)) {
		data1 = closure->data;
		data2 = param_values->data[0].v_pointer;
	} else {
		data1 = param_values->data[0].v_pointer;
		data2 = closure->data;
	}
	callback = ((GMarshalFunc_VOID__STRING_ULONG) ((marshal_data ? marshal_data : cc->callback)));
	callback (data1, g_value_get_string (param_values + 1), g_value_get_ulong (param_values + 2), data2);
}



