/*
 *  tuntun-auth-dialog.vala is a part of Tuntun
 *
 *  Tuntun: a simple applet to manage OpenVPN connections
 *
 *  Copyright (C) 2008  Andrea Del Signore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 *  Author:
 *     Andrea Del Signore <sejerpz@tin.it>       
 */

#include "tuntun-auth-dialog.h"
#include <config.h>
#include <glib/gi18n-lib.h>
#include <gnome-keyring.h>
#include "tuntun-utils.h"
#include "tuntun-connection-info.h"




struct _TuntunAuthDialogPrivate {
	TuntunConnection* _connection;
	TuntunAuthenticationModes _mode;
	char* _type;
	GtkEntry* _entry_user;
	GtkEntry* _entry_pass;
	GtkCheckButton* _checkbutton_save_in_keyring;
	guint _keyring_item_id;
	TuntunAuthSteps _current;
};

#define TUNTUN_AUTH_DIALOG_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), TUNTUN_TYPE_AUTH_DIALOG, TuntunAuthDialogPrivate))
enum  {
	TUNTUN_AUTH_DIALOG_DUMMY_PROPERTY,
	TUNTUN_AUTH_DIALOG_CONNECTION,
	TUNTUN_AUTH_DIALOG_MODE,
	TUNTUN_AUTH_DIALOG_TYPE
};
static void _tuntun_auth_dialog_on_button_ok_clicked_gtk_button_clicked (GtkButton* _sender, gpointer self);
static void _tuntun_auth_dialog_on_button_cancel_clicked_gtk_button_clicked (GtkButton* _sender, gpointer self);
static void tuntun_auth_dialog_initialize_ui (TuntunAuthDialog* self);
static void tuntun_auth_dialog_on_button_ok_clicked (TuntunAuthDialog* self, GtkButton* sender);
static void tuntun_auth_dialog_on_button_cancel_clicked (TuntunAuthDialog* self, GtkButton* sender);
static void _tuntun_auth_dialog_on_item_delete_done_gnome_keyring_operation_done_callback (GnomeKeyringResult result, gpointer self);
static void tuntun_auth_dialog_on_connection_authentication_failed (TuntunAuthDialog* self, TuntunConnection* connection, TuntunAuthenticationModes mode, const char* type);
static void tuntun_auth_dialog_authenticate_keyring (TuntunAuthDialog* self);
static void tuntun_auth_dialog_authenticate_dialog (TuntunAuthDialog* self);
static void tuntun_auth_dialog_authenticate_dialog_finish (TuntunAuthDialog* self, gboolean cancel);
static void _tuntun_auth_dialog_on_find_password_done_gnome_keyring_operation_get_list_callback (GnomeKeyringResult result, GList* list, gpointer self);
static void tuntun_auth_dialog_authenticate_keyring_auth (TuntunAuthDialog* self);
static void tuntun_auth_dialog_authenticate_keyring_private_key (TuntunAuthDialog* self);
static void tuntun_auth_dialog_on_find_password_done (TuntunAuthDialog* self, GnomeKeyringResult result, GList* list);
static void tuntun_auth_dialog_on_item_delete_done (TuntunAuthDialog* self, GnomeKeyringResult result);
static void tuntun_auth_dialog_on_set_password_done (TuntunAuthDialog* self, GnomeKeyringResult result, guint val);
static void _tuntun_auth_dialog_on_set_password_done_gnome_keyring_operation_get_int_callback (GnomeKeyringResult result, guint val, gpointer self);
static void tuntun_auth_dialog_save (TuntunAuthDialog* self, const char* username, const char* password);
static void tuntun_auth_dialog_save_private_key_password (TuntunAuthDialog* self, const char* private_key_password);
static void tuntun_auth_dialog_set_connection (TuntunAuthDialog* self, TuntunConnection* value);
static void tuntun_auth_dialog_set_mode (TuntunAuthDialog* self, TuntunAuthenticationModes value);
static void tuntun_auth_dialog_set_type (TuntunAuthDialog* self, const char* value);
static void _tuntun_auth_dialog_on_connection_authentication_failed_tuntun_connection_authentication_failed (TuntunConnection* _sender, TuntunAuthenticationModes mode, const char* type, gpointer self);
static GObject * tuntun_auth_dialog_constructor (GType type, guint n_construct_properties, GObjectConstructParam * construct_properties);
static gpointer tuntun_auth_dialog_parent_class = NULL;
static void tuntun_auth_dialog_dispose (GObject * obj);
static int _vala_strcmp0 (const char * str1, const char * str2);




GType tuntun_auth_steps_get_type (void) {
	static GType tuntun_auth_steps_type_id = 0;
	if (G_UNLIKELY (tuntun_auth_steps_type_id == 0)) {
		static const GEnumValue values[] = {{TUNTUN_AUTH_STEPS_KEYRING, "TUNTUN_AUTH_STEPS_KEYRING", "keyring"}, {TUNTUN_AUTH_STEPS_DIALOG, "TUNTUN_AUTH_STEPS_DIALOG", "dialog"}, {0, NULL, NULL}};
		tuntun_auth_steps_type_id = g_enum_register_static ("TuntunAuthSteps", values);
	}
	return tuntun_auth_steps_type_id;
}


TuntunAuthDialog* tuntun_auth_dialog_new (TuntunConnection* connection, TuntunAuthenticationModes mode, const char* type) {
	GParameter * __params;
	GParameter * __params_it;
	TuntunAuthDialog * self;
	g_return_val_if_fail (TUNTUN_IS_CONNECTION (connection), NULL);
	g_return_val_if_fail (type != NULL, NULL);
	__params = g_new0 (GParameter, 3);
	__params_it = __params;
	__params_it->name = "connection";
	g_value_init (&__params_it->value, TUNTUN_TYPE_CONNECTION);
	g_value_set_object (&__params_it->value, connection);
	__params_it++;
	__params_it->name = "mode";
	g_value_init (&__params_it->value, TUNTUN_TYPE_AUTHENTICATION_MODES);
	g_value_set_enum (&__params_it->value, mode);
	__params_it++;
	__params_it->name = "type";
	g_value_init (&__params_it->value, G_TYPE_STRING);
	g_value_set_string (&__params_it->value, type);
	__params_it++;
	self = g_object_newv (TUNTUN_TYPE_AUTH_DIALOG, __params_it - __params, __params);
	while (__params_it > __params) {
		--__params_it;
		g_value_unset (&__params_it->value);
	}
	g_free (__params);
	return self;
}


static void _tuntun_auth_dialog_on_button_ok_clicked_gtk_button_clicked (GtkButton* _sender, gpointer self) {
	tuntun_auth_dialog_on_button_ok_clicked (self, _sender);
}


static void _tuntun_auth_dialog_on_button_cancel_clicked_gtk_button_clicked (GtkButton* _sender, gpointer self) {
	tuntun_auth_dialog_on_button_cancel_clicked (self, _sender);
}


static void tuntun_auth_dialog_initialize_ui (TuntunAuthDialog* self) {
	GError * inner_error;
	GtkBuilder* builder;
	GtkVBox* _tmp1;
	GtkVBox* childs;
	GtkButton* _tmp2;
	GtkButton* button;
	GtkButton* _tmp4;
	GtkButton* _tmp3;
	GtkLabel* _tmp5;
	GtkLabel* label;
	GtkLabel* _tmp6;
	GtkLabel* label_message;
	char* label_text;
	TuntunAuthenticationModes _tmp13;
	g_return_if_fail (TUNTUN_IS_AUTH_DIALOG (self));
	inner_error = NULL;
	builder = gtk_builder_new ();
	gtk_builder_set_translation_domain (builder, GETTEXT_PACKAGE);
	{
		char* _tmp0;
		_tmp0 = NULL;
		gtk_builder_add_from_file (builder, (_tmp0 = tuntun_utils_get_ui_path ("tuntun-auth-dialog.ui")), &inner_error);
		if (inner_error != NULL) {
			goto __catch9_g_error;
		}
		_tmp0 = (g_free (_tmp0), NULL);
	}
	goto __finally9;
	__catch9_g_error:
	{
		GError * err;
		err = inner_error;
		inner_error = NULL;
		{
			tuntun_utils_display_error ("initialize_ui", err->message);
			(err == NULL ? NULL : (err = (g_error_free (err), NULL)));
		}
	}
	__finally9:
	;
	_tmp1 = NULL;
	childs = (_tmp1 = GTK_VBOX (gtk_builder_get_object (builder, "vbox_authentication")), (_tmp1 == NULL ? NULL : g_object_ref (_tmp1)));
	g_assert (childs != NULL);
	gtk_container_add (GTK_CONTAINER (self), GTK_WIDGET (childs));
	g_object_set (G_OBJECT (self), "border-width", 12, NULL);
	gtk_window_set_resizable (GTK_WINDOW (self), FALSE);
	_tmp2 = NULL;
	button = (_tmp2 = GTK_BUTTON (gtk_builder_get_object (builder, "button_auth_ok")), (_tmp2 == NULL ? NULL : g_object_ref (_tmp2)));
	g_assert (button != NULL);
	g_signal_connect_object (button, "clicked", ((GCallback) (_tuntun_auth_dialog_on_button_ok_clicked_gtk_button_clicked)), self, 0);
	_tmp4 = NULL;
	_tmp3 = NULL;
	button = (_tmp4 = (_tmp3 = GTK_BUTTON (gtk_builder_get_object (builder, "button_auth_cancel")), (_tmp3 == NULL ? NULL : g_object_ref (_tmp3))), (button == NULL ? NULL : (button = (g_object_unref (button), NULL))), _tmp4);
	g_assert (button != NULL);
	g_signal_connect_object (button, "clicked", ((GCallback) (_tuntun_auth_dialog_on_button_cancel_clicked_gtk_button_clicked)), self, 0);
	self->priv->_entry_user = GTK_ENTRY (gtk_builder_get_object (builder, "entry_username"));
	g_assert (self->priv->_entry_user != NULL);
	self->priv->_entry_pass = GTK_ENTRY (gtk_builder_get_object (builder, "entry_password"));
	g_assert (self->priv->_entry_pass != NULL);
	self->priv->_checkbutton_save_in_keyring = GTK_CHECK_BUTTON (gtk_builder_get_object (builder, "checkbutton_save_in_keyring"));
	g_assert (self->priv->_checkbutton_save_in_keyring != NULL);
	_tmp5 = NULL;
	label = (_tmp5 = GTK_LABEL (gtk_builder_get_object (builder, "label_username")), (_tmp5 == NULL ? NULL : g_object_ref (_tmp5)));
	_tmp6 = NULL;
	label_message = (_tmp6 = GTK_LABEL (gtk_builder_get_object (builder, "label_message")), (_tmp6 == NULL ? NULL : g_object_ref (_tmp6)));
	label_text = NULL;
	g_assert (label != NULL);
	g_assert (label_message != NULL);
	_tmp13 = self->priv->_mode;
	if (_tmp13 == TUNTUN_AUTHENTICATION_MODES_USERNAME_PASSWORD)
	do {
		char* _tmp8;
		const char* _tmp7;
		char* _tmp9;
		_tmp8 = NULL;
		_tmp7 = NULL;
		label_text = (_tmp8 = (_tmp7 = _ ("Username and password for %s"), (_tmp7 == NULL ? NULL : g_strdup (_tmp7))), (label_text = (g_free (label_text), NULL)), _tmp8);
		_tmp9 = NULL;
		gtk_label_set_text (label_message, (_tmp9 = g_strdup_printf (label_text, tuntun_connection_info_get_name (tuntun_connection_get_info (self->priv->_connection)))));
		_tmp9 = (g_free (_tmp9), NULL);
		gtk_widget_show (GTK_WIDGET (self->priv->_entry_user));
		gtk_widget_show (GTK_WIDGET (label));
		break;
	} while (0); else if (_tmp13 == TUNTUN_AUTHENTICATION_MODES_PASSWORD_ONLY)
	do {
		char* _tmp11;
		const char* _tmp10;
		char* _tmp12;
		_tmp11 = NULL;
		_tmp10 = NULL;
		label_text = (_tmp11 = (_tmp10 = _ ("Private key password for %s"), (_tmp10 == NULL ? NULL : g_strdup (_tmp10))), (label_text = (g_free (label_text), NULL)), _tmp11);
		_tmp12 = NULL;
		gtk_label_set_text (label_message, (_tmp12 = g_strdup_printf (label_text, tuntun_connection_info_get_name (tuntun_connection_get_info (self->priv->_connection)))));
		_tmp12 = (g_free (_tmp12), NULL);
		gtk_widget_hide (GTK_WIDGET (self->priv->_entry_user));
		gtk_widget_hide (GTK_WIDGET (label));
		break;
	} while (0);
	gtk_window_set_position (GTK_WINDOW (self), GTK_WIN_POS_CENTER);
	(builder == NULL ? NULL : (builder = (g_object_unref (builder), NULL)));
	(childs == NULL ? NULL : (childs = (g_object_unref (childs), NULL)));
	(button == NULL ? NULL : (button = (g_object_unref (button), NULL)));
	(label == NULL ? NULL : (label = (g_object_unref (label), NULL)));
	(label_message == NULL ? NULL : (label_message = (g_object_unref (label_message), NULL)));
	label_text = (g_free (label_text), NULL);
}


static void tuntun_auth_dialog_on_button_ok_clicked (TuntunAuthDialog* self, GtkButton* sender) {
	g_return_if_fail (TUNTUN_IS_AUTH_DIALOG (self));
	g_return_if_fail (GTK_IS_BUTTON (sender));
	tuntun_auth_dialog_authenticate_dialog_finish (self, FALSE);
}


static void tuntun_auth_dialog_on_button_cancel_clicked (TuntunAuthDialog* self, GtkButton* sender) {
	g_return_if_fail (TUNTUN_IS_AUTH_DIALOG (self));
	g_return_if_fail (GTK_IS_BUTTON (sender));
	tuntun_auth_dialog_authenticate_dialog_finish (self, TRUE);
	gtk_object_destroy (GTK_OBJECT (self));
}


static void _tuntun_auth_dialog_on_item_delete_done_gnome_keyring_operation_done_callback (GnomeKeyringResult result, gpointer self) {
	tuntun_auth_dialog_on_item_delete_done (self, result);
}


static void tuntun_auth_dialog_on_connection_authentication_failed (TuntunAuthDialog* self, TuntunConnection* connection, TuntunAuthenticationModes mode, const char* type) {
	g_return_if_fail (TUNTUN_IS_AUTH_DIALOG (self));
	g_return_if_fail (TUNTUN_IS_CONNECTION (connection));
	g_return_if_fail (type != NULL);
	/*delete bogus keyring item*/
	if (self->priv->_keyring_item_id != 0) {
		gnome_keyring_item_delete (NULL, self->priv->_keyring_item_id, _tuntun_auth_dialog_on_item_delete_done_gnome_keyring_operation_done_callback, self, NULL);
	}
}


void tuntun_auth_dialog_authenticate (TuntunAuthDialog* self) {
	g_return_if_fail (TUNTUN_IS_AUTH_DIALOG (self));
	self->priv->_current = TUNTUN_AUTH_STEPS_KEYRING;
	tuntun_auth_dialog_authenticate_keyring (self);
}


static void tuntun_auth_dialog_authenticate_keyring (TuntunAuthDialog* self) {
	TuntunAuthenticationModes _tmp0;
	g_return_if_fail (TUNTUN_IS_AUTH_DIALOG (self));
	_tmp0 = self->priv->_mode;
	if (_tmp0 == TUNTUN_AUTHENTICATION_MODES_USERNAME_PASSWORD)
	do {
		tuntun_auth_dialog_authenticate_keyring_auth (self);
		break;
	} while (0); else if (_tmp0 == TUNTUN_AUTHENTICATION_MODES_PASSWORD_ONLY)
	do {
		tuntun_auth_dialog_authenticate_keyring_private_key (self);
		break;
	} while (0);
}


static void tuntun_auth_dialog_authenticate_dialog (TuntunAuthDialog* self) {
	g_return_if_fail (TUNTUN_IS_AUTH_DIALOG (self));
	tuntun_auth_dialog_initialize_ui (self);
	self->priv->_current = TUNTUN_AUTH_STEPS_DIALOG;
	gtk_widget_show_all (GTK_WIDGET (self));
}


static void tuntun_auth_dialog_authenticate_dialog_finish (TuntunAuthDialog* self, gboolean cancel) {
	char* username;
	char* password;
	TuntunAuthenticationModes _tmp6;
	g_return_if_fail (TUNTUN_IS_AUTH_DIALOG (self));
	username = NULL;
	password = NULL;
	_tmp6 = self->priv->_mode;
	if (_tmp6 == TUNTUN_AUTHENTICATION_MODES_USERNAME_PASSWORD)
	do {
		char* _tmp1;
		const char* _tmp0;
		char* _tmp3;
		const char* _tmp2;
		_tmp1 = NULL;
		_tmp0 = NULL;
		username = (_tmp1 = (_tmp0 = gtk_entry_get_text (self->priv->_entry_user), (_tmp0 == NULL ? NULL : g_strdup (_tmp0))), (username = (g_free (username), NULL)), _tmp1);
		_tmp3 = NULL;
		_tmp2 = NULL;
		password = (_tmp3 = (_tmp2 = gtk_entry_get_text (self->priv->_entry_pass), (_tmp2 == NULL ? NULL : g_strdup (_tmp2))), (password = (g_free (password), NULL)), _tmp3);
		break;
	} while (0); else if (_tmp6 == TUNTUN_AUTHENTICATION_MODES_PASSWORD_ONLY)
	do {
		char* _tmp5;
		const char* _tmp4;
		_tmp5 = NULL;
		_tmp4 = NULL;
		password = (_tmp5 = (_tmp4 = gtk_entry_get_text (self->priv->_entry_pass), (_tmp4 == NULL ? NULL : g_strdup (_tmp4))), (password = (g_free (password), NULL)), _tmp5);
		break;
	} while (0);
	if (!cancel) {
		tuntun_connection_authenticate (self->priv->_connection, self->priv->_type, username, password);
		if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (self->priv->_checkbutton_save_in_keyring))) {
			if (username != NULL) {
				tuntun_auth_dialog_save (self, username, password);
			} else {
				tuntun_auth_dialog_save_private_key_password (self, password);
			}
		} else {
			gtk_object_destroy (GTK_OBJECT (self));
		}
	} else {
		tuntun_connection_cancel_authentication (self->priv->_connection, self->priv->_type, username);
	}
	username = (g_free (username), NULL);
	password = (g_free (password), NULL);
}


static void _tuntun_auth_dialog_on_find_password_done_gnome_keyring_operation_get_list_callback (GnomeKeyringResult result, GList* list, gpointer self) {
	tuntun_auth_dialog_on_find_password_done (self, result, list);
}


static void tuntun_auth_dialog_authenticate_keyring_auth (TuntunAuthDialog* self) {
	g_return_if_fail (TUNTUN_IS_AUTH_DIALOG (self));
	gnome_keyring_find_network_password (NULL, NULL, tuntun_connection_info_get_address (tuntun_connection_get_info (self->priv->_connection)), NULL, "vpn", "auth", ((guint) (0)), _tuntun_auth_dialog_on_find_password_done_gnome_keyring_operation_get_list_callback, self, NULL);
}


static void tuntun_auth_dialog_authenticate_keyring_private_key (TuntunAuthDialog* self) {
	g_return_if_fail (TUNTUN_IS_AUTH_DIALOG (self));
	gnome_keyring_find_network_password (NULL, NULL, tuntun_connection_info_get_address (tuntun_connection_get_info (self->priv->_connection)), NULL, "vpn", "private-key", ((guint) (0)), _tuntun_auth_dialog_on_find_password_done_gnome_keyring_operation_get_list_callback, self, NULL);
}


static void tuntun_auth_dialog_on_find_password_done (TuntunAuthDialog* self, GnomeKeyringResult result, GList* list) {
	char* username;
	char* password;
	g_return_if_fail (TUNTUN_IS_AUTH_DIALOG (self));
	username = NULL;
	password = NULL;
	if (result == GNOME_KEYRING_RESULT_OK && g_list_length (list) > 0) {
		GnomeKeyringNetworkPasswordData* npd;
		npd = ((GnomeKeyringNetworkPasswordData*) (g_list_first (list)->data));
		self->priv->_keyring_item_id = npd->item_id;
		if (_vala_strcmp0 (npd->authtype, "auth") == 0) {
			char* _tmp1;
			const char* _tmp0;
			char* _tmp3;
			const char* _tmp2;
			_tmp1 = NULL;
			_tmp0 = NULL;
			username = (_tmp1 = (_tmp0 = npd->user, (_tmp0 == NULL ? NULL : g_strdup (_tmp0))), (username = (g_free (username), NULL)), _tmp1);
			_tmp3 = NULL;
			_tmp2 = NULL;
			password = (_tmp3 = (_tmp2 = npd->password, (_tmp2 == NULL ? NULL : g_strdup (_tmp2))), (password = (g_free (password), NULL)), _tmp3);
		} else {
			if (_vala_strcmp0 (npd->authtype, "private-key") == 0) {
				char* _tmp4;
				char* _tmp6;
				const char* _tmp5;
				_tmp4 = NULL;
				username = (_tmp4 = NULL, (username = (g_free (username), NULL)), _tmp4);
				_tmp6 = NULL;
				_tmp5 = NULL;
				password = (_tmp6 = (_tmp5 = npd->password, (_tmp5 == NULL ? NULL : g_strdup (_tmp5))), (password = (g_free (password), NULL)), _tmp6);
			}
		}
		tuntun_connection_authenticate (self->priv->_connection, self->priv->_type, username, password);
	} else {
		tuntun_auth_dialog_authenticate_dialog (self);
	}
	username = (g_free (username), NULL);
	password = (g_free (password), NULL);
}


static void tuntun_auth_dialog_on_item_delete_done (TuntunAuthDialog* self, GnomeKeyringResult result) {
	g_return_if_fail (TUNTUN_IS_AUTH_DIALOG (self));
	if (result != GNOME_KEYRING_RESULT_OK) {
		tuntun_utils_display_error (_ ("Keyring"), _ ("Error deleting credentials for '%s'"));
	}
}


static void tuntun_auth_dialog_on_set_password_done (TuntunAuthDialog* self, GnomeKeyringResult result, guint val) {
	g_return_if_fail (TUNTUN_IS_AUTH_DIALOG (self));
	if (result != GNOME_KEYRING_RESULT_OK) {
		tuntun_utils_display_error (_ ("Keyring"), _ ("Error saving credentials for '%s'"));
	}
	gtk_object_destroy (GTK_OBJECT (self));
}


static void _tuntun_auth_dialog_on_set_password_done_gnome_keyring_operation_get_int_callback (GnomeKeyringResult result, guint val, gpointer self) {
	tuntun_auth_dialog_on_set_password_done (self, result, val);
}


static void tuntun_auth_dialog_save (TuntunAuthDialog* self, const char* username, const char* password) {
	g_return_if_fail (TUNTUN_IS_AUTH_DIALOG (self));
	g_return_if_fail (username != NULL);
	g_return_if_fail (_vala_strcmp0 (username, "") != 0);
	gnome_keyring_set_network_password (NULL, username, NULL, tuntun_connection_info_get_address (tuntun_connection_get_info (self->priv->_connection)), NULL, "vpn", "auth", ((guint) (0)), password, _tuntun_auth_dialog_on_set_password_done_gnome_keyring_operation_get_int_callback, self, NULL);
}


static void tuntun_auth_dialog_save_private_key_password (TuntunAuthDialog* self, const char* private_key_password) {
	g_return_if_fail (TUNTUN_IS_AUTH_DIALOG (self));
	g_return_if_fail (private_key_password != NULL);
	gnome_keyring_set_network_password (NULL, NULL, NULL, tuntun_connection_info_get_address (tuntun_connection_get_info (self->priv->_connection)), NULL, "vpn", "private-key", ((guint) (0)), private_key_password, _tuntun_auth_dialog_on_set_password_done_gnome_keyring_operation_get_int_callback, self, NULL);
}


static void tuntun_auth_dialog_set_connection (TuntunAuthDialog* self, TuntunConnection* value) {
	TuntunConnection* _tmp1;
	TuntunConnection* _tmp0;
	g_return_if_fail (TUNTUN_IS_AUTH_DIALOG (self));
	_tmp1 = NULL;
	_tmp0 = NULL;
	self->priv->_connection = (_tmp1 = (_tmp0 = value, (_tmp0 == NULL ? NULL : g_object_ref (_tmp0))), (self->priv->_connection == NULL ? NULL : (self->priv->_connection = (g_object_unref (self->priv->_connection), NULL))), _tmp1);
	g_object_notify (((GObject *) (self)), "connection");
}


static void tuntun_auth_dialog_set_mode (TuntunAuthDialog* self, TuntunAuthenticationModes value) {
	g_return_if_fail (TUNTUN_IS_AUTH_DIALOG (self));
	self->priv->_mode = value;
	g_object_notify (((GObject *) (self)), "mode");
}


static void tuntun_auth_dialog_set_type (TuntunAuthDialog* self, const char* value) {
	char* _tmp1;
	const char* _tmp0;
	g_return_if_fail (TUNTUN_IS_AUTH_DIALOG (self));
	_tmp1 = NULL;
	_tmp0 = NULL;
	self->priv->_type = (_tmp1 = (_tmp0 = value, (_tmp0 == NULL ? NULL : g_strdup (_tmp0))), (self->priv->_type = (g_free (self->priv->_type), NULL)), _tmp1);
	g_object_notify (((GObject *) (self)), "type");
}


static void _tuntun_auth_dialog_on_connection_authentication_failed_tuntun_connection_authentication_failed (TuntunConnection* _sender, TuntunAuthenticationModes mode, const char* type, gpointer self) {
	tuntun_auth_dialog_on_connection_authentication_failed (self, _sender, mode, type);
}


static GObject * tuntun_auth_dialog_constructor (GType type, guint n_construct_properties, GObjectConstructParam * construct_properties) {
	GObject * obj;
	TuntunAuthDialogClass * klass;
	GObjectClass * parent_class;
	TuntunAuthDialog * self;
	klass = TUNTUN_AUTH_DIALOG_CLASS (g_type_class_peek (TUNTUN_TYPE_AUTH_DIALOG));
	parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
	obj = parent_class->constructor (type, n_construct_properties, construct_properties);
	self = TUNTUN_AUTH_DIALOG (obj);
	{
		g_signal_connect_object (self->priv->_connection, "authentication-failed", ((GCallback) (_tuntun_auth_dialog_on_connection_authentication_failed_tuntun_connection_authentication_failed)), self, 0);
	}
	return obj;
}


static void tuntun_auth_dialog_set_property (GObject * object, guint property_id, const GValue * value, GParamSpec * pspec) {
	TuntunAuthDialog * self;
	self = TUNTUN_AUTH_DIALOG (object);
	switch (property_id) {
		case TUNTUN_AUTH_DIALOG_CONNECTION:
		tuntun_auth_dialog_set_connection (self, g_value_get_object (value));
		break;
		case TUNTUN_AUTH_DIALOG_MODE:
		tuntun_auth_dialog_set_mode (self, g_value_get_enum (value));
		break;
		case TUNTUN_AUTH_DIALOG_TYPE:
		tuntun_auth_dialog_set_type (self, g_value_get_string (value));
		break;
		default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}


static void tuntun_auth_dialog_class_init (TuntunAuthDialogClass * klass) {
	tuntun_auth_dialog_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (TuntunAuthDialogPrivate));
	G_OBJECT_CLASS (klass)->set_property = tuntun_auth_dialog_set_property;
	G_OBJECT_CLASS (klass)->constructor = tuntun_auth_dialog_constructor;
	G_OBJECT_CLASS (klass)->dispose = tuntun_auth_dialog_dispose;
	g_object_class_install_property (G_OBJECT_CLASS (klass), TUNTUN_AUTH_DIALOG_CONNECTION, g_param_spec_object ("connection", "connection", "connection", TUNTUN_TYPE_CONNECTION, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (G_OBJECT_CLASS (klass), TUNTUN_AUTH_DIALOG_MODE, g_param_spec_enum ("mode", "mode", "mode", TUNTUN_TYPE_AUTHENTICATION_MODES, 0, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (G_OBJECT_CLASS (klass), TUNTUN_AUTH_DIALOG_TYPE, g_param_spec_string ("type", "type", "type", NULL, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
}


static void tuntun_auth_dialog_instance_init (TuntunAuthDialog * self) {
	self->priv = TUNTUN_AUTH_DIALOG_GET_PRIVATE (self);
	self->priv->_connection = NULL;
	self->priv->_mode = TUNTUN_AUTHENTICATION_MODES_USERNAME_PASSWORD;
	self->priv->_type = NULL;
	self->priv->_keyring_item_id = ((guint) (0));
	self->priv->_current = TUNTUN_AUTH_STEPS_KEYRING;
}


static void tuntun_auth_dialog_dispose (GObject * obj) {
	TuntunAuthDialog * self;
	self = TUNTUN_AUTH_DIALOG (obj);
	(self->priv->_connection == NULL ? NULL : (self->priv->_connection = (g_object_unref (self->priv->_connection), NULL)));
	self->priv->_type = (g_free (self->priv->_type), NULL);
	G_OBJECT_CLASS (tuntun_auth_dialog_parent_class)->dispose (obj);
}


GType tuntun_auth_dialog_get_type (void) {
	static volatile gsize tuntun_auth_dialog_type_id = 0;
	if (g_once_init_enter (&tuntun_auth_dialog_type_id)) {
		GType tuntun_auth_dialog_type_id_temp;
		static const GTypeInfo g_define_type_info = { sizeof (TuntunAuthDialogClass), (GBaseInitFunc) NULL, (GBaseFinalizeFunc) NULL, (GClassInitFunc) tuntun_auth_dialog_class_init, (GClassFinalizeFunc) NULL, NULL, sizeof (TuntunAuthDialog), 0, (GInstanceInitFunc) tuntun_auth_dialog_instance_init };
		tuntun_auth_dialog_type_id_temp = g_type_register_static (GTK_TYPE_WINDOW, "TuntunAuthDialog", &g_define_type_info, 0);
		g_once_init_leave (&tuntun_auth_dialog_type_id, tuntun_auth_dialog_type_id_temp);
	}
	return tuntun_auth_dialog_type_id;
}


static int _vala_strcmp0 (const char * str1, const char * str2) {
	if (str1 == NULL) {
		return -(str1 != str2);
	}
	if (str2 == NULL) {
		return (str1 != str2);
	}
	return strcmp (str1, str2);
}




