/*
 *  tuntun-connections.vala is a part of Tuntun
 *
 *  Tuntun: a simple applet to manage OpenVPN connections
 *
 *  Copyright (C) 2008  Andrea Del Signore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 *  Author:
 *     Andrea Del Signore <sejerpz@tin.it>       
 */

#include "tuntun-connections.h"
#include <glib/gstdio.h>
#include "tuntun-utils.h"
#include "tuntun-constants.h"
#include "tuntun-connection-info.h"




struct _TuntunConnectionsPrivate {
	char* _config_file;
	guint _timeout_id;
};

#define TUNTUN_CONNECTIONS_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), TUNTUN_TYPE_CONNECTIONS, TuntunConnectionsPrivate))
enum  {
	TUNTUN_CONNECTIONS_DUMMY_PROPERTY
};
static void _g_list_free_g_object_unref (GList* self);
static void tuntun_connections_load (TuntunConnections* self);
static void _tuntun_connections_xmlconfig_start_element_handler_gmarkup_parser_start_element_func (GMarkupParseContext* context, const char* element_name, char** attribute_names, char** attribute_values, void* user_data);
static void _tuntun_connections_xmlconfig_end_element_handler_gmarkup_parser_end_element_func (GMarkupParseContext* context, const char* element_name, void* user_data);
static void _tuntun_connections_xmlconfig_text_handler_gmarkup_parser_text_func (GMarkupParseContext* context, const char* text, gulong text_len, void* user_data);
static void tuntun_connections_load_config (TuntunConnections* self, const char* filename, GError** error);
static void tuntun_connections_save_config (TuntunConnections* self, const char* filename, GError** error);
static void tuntun_connections_xmlconfig_start_element_handler (GMarkupParseContext* context, const char* element_name, char** attribute_names, char** attribute_values, void* user_data);
static void _tuntun_connections_on_connection_authentication_required_tuntun_connection_authentication_required (TuntunConnection* _sender, TuntunAuthenticationModes mode, const char* type, gpointer self);
static void _tuntun_connections_on_connection_authentication_failed_tuntun_connection_authentication_failed (TuntunConnection* _sender, TuntunAuthenticationModes mode, const char* type, gpointer self);
static void _tuntun_connections_on_connection_fatal_error_tuntun_connection_control_channel_fatal_error (TuntunConnection* _sender, const char* error, gpointer self);
static void _tuntun_connections_on_connection_data_received_tuntun_connection_control_channel_data_received (TuntunConnection* _sender, const char* data, gpointer self);
static void tuntun_connections_on_connection_data_received (TuntunConnections* self, TuntunConnection* connection, const char* data);
static void tuntun_connections_real_on_connection_activity (TuntunConnections* self, TuntunConnection* connection);
static void tuntun_connections_real_on_connection_fatal_error (TuntunConnections* self, TuntunConnection* connection, const char* error);
static void tuntun_connections_real_on_connection_authentication_required (TuntunConnections* self, TuntunConnection* connection, TuntunAuthenticationModes mode, const char* type);
static void tuntun_connections_real_on_connection_authentication_failed (TuntunConnections* self, TuntunConnection* connection, TuntunAuthenticationModes mode, const char* type);
static void tuntun_connections_real_on_connection_removed (TuntunConnections* self, TuntunConnection* connection);
static void tuntun_connections_real_on_connection_added (TuntunConnections* self, TuntunConnection* connection);
static gboolean _tuntun_connections_connection_recheck_gsource_func (gpointer self);
static void tuntun_connections_real_on_connection_status_changed (TuntunConnections* self, TuntunConnection* connection);
static gboolean tuntun_connections_connection_recheck (TuntunConnections* self);
static void tuntun_connections_connection_notify_property_changed (TuntunConnection* connection, GParamSpec* param, TuntunConnections* connections);
static void tuntun_connections_xmlconfig_end_element_handler (GMarkupParseContext* context, const char* element_name, void* user_data);
static void tuntun_connections_xmlconfig_text_handler (GMarkupParseContext* context, const char* text, gulong text_len, void* user_data);
static GObject * tuntun_connections_constructor (GType type, guint n_construct_properties, GObjectConstructParam * construct_properties);
static gpointer tuntun_connections_parent_class = NULL;
static void tuntun_connections_dispose (GObject * obj);
static int _vala_strcmp0 (const char * str1, const char * str2);


static void g_cclosure_user_marshal_VOID__OBJECT_STRING (GClosure * closure, GValue * return_value, guint n_param_values, const GValue * param_values, gpointer invocation_hint, gpointer marshal_data);
static void g_cclosure_user_marshal_VOID__OBJECT_ENUM_STRING (GClosure * closure, GValue * return_value, guint n_param_values, const GValue * param_values, gpointer invocation_hint, gpointer marshal_data);

static void _g_list_free_g_object_unref (GList* self) {
	g_list_foreach (self, ((GFunc) (g_object_unref)), NULL);
	g_list_free (self);
}


static void tuntun_connections_load (TuntunConnections* self) {
	GError * inner_error;
	char* _tmp0;
	g_return_if_fail (TUNTUN_IS_CONNECTIONS (self));
	inner_error = NULL;
	_tmp0 = NULL;
	self->priv->_config_file = (_tmp0 = tuntun_utils_gnome_util_home_file (TUNTUN_CONSTANTS_CONNECTIONS_FILENAME), (self->priv->_config_file = (g_free (self->priv->_config_file), NULL)), _tmp0);
	{
		tuntun_connections_load_config (self, self->priv->_config_file, &inner_error);
		if (inner_error != NULL) {
			goto __catch3_g_error;
		}
	}
	goto __finally3;
	__catch3_g_error:
	{
		GError * err;
		err = inner_error;
		inner_error = NULL;
		{
			g_warning ("tuntun-connections.vala:59: Error reading config file: %s, %s", self->priv->_config_file, err->message);
			(err == NULL ? NULL : (err = (g_error_free (err), NULL)));
		}
	}
	__finally3:
	;
}


static void _tuntun_connections_xmlconfig_start_element_handler_gmarkup_parser_start_element_func (GMarkupParseContext* context, const char* element_name, char** attribute_names, char** attribute_values, void* user_data) {
	tuntun_connections_xmlconfig_start_element_handler (context, element_name, attribute_names, attribute_values, user_data);
}


static void _tuntun_connections_xmlconfig_end_element_handler_gmarkup_parser_end_element_func (GMarkupParseContext* context, const char* element_name, void* user_data) {
	tuntun_connections_xmlconfig_end_element_handler (context, element_name, user_data);
}


static void _tuntun_connections_xmlconfig_text_handler_gmarkup_parser_text_func (GMarkupParseContext* context, const char* text, gulong text_len, void* user_data) {
	tuntun_connections_xmlconfig_text_handler (context, text, text_len, user_data);
}


static void tuntun_connections_load_config (TuntunConnections* self, const char* filename, GError** error) {
	GError * inner_error;
	g_return_if_fail (TUNTUN_IS_CONNECTIONS (self));
	g_return_if_fail (filename != NULL);
	inner_error = NULL;
	{
		GMarkupParser parser = {0};
		char* content;
		glong len;
		char* _tmp2;
		gboolean _tmp1;
		char* _tmp0;
		GMarkupParseContext* context;
		content = NULL;
		len = 0L;
		if (!g_file_test (filename, G_FILE_TEST_EXISTS)) {
			content = (g_free (content), NULL);
			return;
		}
		/*no file*/
		if (!(g_file_test (filename, G_FILE_TEST_IS_REGULAR) || g_file_test (filename, G_FILE_TEST_IS_SYMLINK))) {
			inner_error = g_error_new (g_quark_from_string ("File exists, but not a regular file or link"), 0, "File exists, but not a regular file or link", NULL);
			if (inner_error != NULL) {
				goto __catch4_g_error;
			}
		}
		/* read config file*/
		_tmp2 = NULL;
		_tmp0 = NULL;
		_tmp1 = g_file_get_contents (filename, &_tmp0, &len, &inner_error);
		content = (_tmp2 = _tmp0, (content = (g_free (content), NULL)), _tmp2);
		_tmp1;
		if (inner_error != NULL) {
			goto __catch4_g_error;
		}
		parser.start_element = _tuntun_connections_xmlconfig_start_element_handler_gmarkup_parser_start_element_func;
		parser.end_element = _tuntun_connections_xmlconfig_end_element_handler_gmarkup_parser_end_element_func;
		parser.text = _tuntun_connections_xmlconfig_text_handler_gmarkup_parser_text_func;
		context = g_markup_parse_context_new (&parser, G_MARKUP_TREAT_CDATA_AS_TEXT, self, NULL);
		g_markup_parse_context_parse (context, content, len, &inner_error);
		if (inner_error != NULL) {
			goto __catch4_g_error;
		}
		content = (g_free (content), NULL);
		(context == NULL ? NULL : (context = (g_markup_parse_context_free (context), NULL)));
	}
	goto __finally4;
	__catch4_g_error:
	{
		GError * err;
		err = inner_error;
		inner_error = NULL;
		{
			GError* _tmp3;
			_tmp3 = NULL;
			inner_error = (_tmp3 = err, (_tmp3 == NULL || g_error_copy == NULL ? ((gpointer) (_tmp3)) : g_error_copy (((gpointer) (_tmp3)))));
			if (inner_error != NULL) {
				g_propagate_error (error, inner_error);
				(err == NULL ? NULL : (err = (g_error_free (err), NULL)));
				return;
			}
			(err == NULL ? NULL : (err = (g_error_free (err), NULL)));
		}
	}
	__finally4:
	;
}


void tuntun_connections_save (TuntunConnections* self, GError** error) {
	GError * inner_error;
	g_return_if_fail (TUNTUN_IS_CONNECTIONS (self));
	inner_error = NULL;
	tuntun_connections_save_config (self, self->priv->_config_file, &inner_error);
	if (inner_error != NULL) {
		g_propagate_error (error, inner_error);
		return;
	}
}


static void tuntun_connections_save_config (TuntunConnections* self, const char* filename, GError** error) {
	GError * inner_error;
	char* content;
	char* _tmp0;
	char* _tmp3;
	g_return_if_fail (TUNTUN_IS_CONNECTIONS (self));
	g_return_if_fail (filename != NULL);
	inner_error = NULL;
	content = NULL;
	_tmp0 = NULL;
	content = (_tmp0 = g_strdup ("<?xml version='1.0'?>\n<connections>\n"), (content = (g_free (content), NULL)), _tmp0);
	{
		GList* connection_collection;
		GList* connection_it;
		connection_collection = self->items;
		for (connection_it = connection_collection; connection_it != NULL; connection_it = connection_it->next) {
			TuntunConnection* _tmp2;
			TuntunConnection* connection;
			_tmp2 = NULL;
			connection = (_tmp2 = ((TuntunConnection*) (connection_it->data)), (_tmp2 == NULL ? NULL : g_object_ref (_tmp2)));
			{
				char* tmp;
				char* _tmp1;
				tmp = g_markup_printf_escaped ("\t<connection \n\t\tname=\"%s\" \n\t\thost=\"%s\" \n\t\tport=\"%d\" \n\t\tquick-connect=\"%s\" >\n\t</connection>\n", tuntun_connection_info_get_name (tuntun_connection_get_info (connection)), tuntun_connection_info_get_address (tuntun_connection_get_info (connection)), tuntun_connection_info_get_port (tuntun_connection_get_info (connection)), (tuntun_connection_info_get_quick_connect (tuntun_connection_get_info (connection)) ? "1" : "0"));
				_tmp1 = NULL;
				content = (_tmp1 = g_strconcat (content, (tmp), NULL), (content = (g_free (content), NULL)), _tmp1);
				(connection == NULL ? NULL : (connection = (g_object_unref (connection), NULL)));
				tmp = (g_free (tmp), NULL);
			}
		}
	}
	_tmp3 = NULL;
	content = (_tmp3 = g_strconcat (content, ("</connections>\n"), NULL), (content = (g_free (content), NULL)), _tmp3);
	{
		g_file_set_contents (filename, content, -1, &inner_error);
		if (inner_error != NULL) {
			if (inner_error->domain == G_FILE_ERROR) {
				goto __catch5_g_file_error;
			}
			g_critical ("file %s: line %d: uncaught error: %s", __FILE__, __LINE__, inner_error->message);
			g_clear_error (&inner_error);
		}
	}
	goto __finally5;
	__catch5_g_file_error:
	{
		GError * err;
		err = inner_error;
		inner_error = NULL;
		{
			g_warning ("tuntun-connections.vala:115: error saving config: %s", err->message);
			(err == NULL ? NULL : (err = (g_error_free (err), NULL)));
		}
	}
	__finally5:
	;
	content = (g_free (content), NULL);
}


static void tuntun_connections_xmlconfig_start_element_handler (GMarkupParseContext* context, const char* element_name, char** attribute_names, char** attribute_values, void* user_data) {
	TuntunConnections* _tmp0;
	TuntunConnections* connections;
	g_return_if_fail (context != NULL);
	g_return_if_fail (element_name != NULL);
	_tmp0 = NULL;
	connections = (_tmp0 = TUNTUN_CONNECTIONS (user_data), (_tmp0 == NULL ? NULL : g_object_ref (_tmp0)));
	if (element_name != NULL && g_utf8_collate ("connection", element_name) == 0) {
		TuntunConnectionInfo* info;
		info = tuntun_connection_info_new ();
		{
			gint i;
			i = 0;
			for (; attribute_names[i] != NULL; i++) {
				if (attribute_names[i] == NULL) {
					continue;
				}
				if (g_utf8_collate ("name", attribute_names[i]) == 0) {
					tuntun_connection_info_set_name (info, attribute_values[i]);
				} else {
					if (g_utf8_collate ("host", attribute_names[i]) == 0) {
						tuntun_connection_info_set_address (info, attribute_values[i]);
					} else {
						if (g_utf8_collate ("port", attribute_names[i]) == 0) {
							tuntun_connection_info_set_port (info, atoi (attribute_values[i]));
						} else {
							if (g_utf8_collate ("quick-connect", attribute_names[i]) == 0) {
								tuntun_connection_info_set_quick_connect (info, ((atoi (attribute_values[i]) != 0 ? TRUE : FALSE)));
							}
						}
					}
				}
			}
		}
		if (tuntun_connection_info_get_name (info) != NULL && _vala_strcmp0 (tuntun_connection_info_get_name (info), "") != 0) {
			TuntunConnection* connection;
			connection = tuntun_connection_new (info);
			tuntun_connections_add (connections, connection);
			(connection == NULL ? NULL : (connection = (g_object_unref (connection), NULL)));
		}
		(info == NULL ? NULL : (info = (g_object_unref (info), NULL)));
	}
	(connections == NULL ? NULL : (connections = (g_object_unref (connections), NULL)));
}


static void _tuntun_connections_on_connection_authentication_required_tuntun_connection_authentication_required (TuntunConnection* _sender, TuntunAuthenticationModes mode, const char* type, gpointer self) {
	tuntun_connections_on_connection_authentication_required (self, _sender, mode, type);
}


static void _tuntun_connections_on_connection_authentication_failed_tuntun_connection_authentication_failed (TuntunConnection* _sender, TuntunAuthenticationModes mode, const char* type, gpointer self) {
	tuntun_connections_on_connection_authentication_failed (self, _sender, mode, type);
}


static void _tuntun_connections_on_connection_fatal_error_tuntun_connection_control_channel_fatal_error (TuntunConnection* _sender, const char* error, gpointer self) {
	tuntun_connections_on_connection_fatal_error (self, _sender, error);
}


static void _tuntun_connections_on_connection_data_received_tuntun_connection_control_channel_data_received (TuntunConnection* _sender, const char* data, gpointer self) {
	tuntun_connections_on_connection_data_received (self, _sender, data);
}


void tuntun_connections_add (TuntunConnections* self, TuntunConnection* connection) {
	TuntunConnection* _tmp0;
	g_return_if_fail (TUNTUN_IS_CONNECTIONS (self));
	g_return_if_fail (TUNTUN_IS_CONNECTION (connection));
	tuntun_connection_initialize (connection);
	_tmp0 = NULL;
	self->items = g_list_append (self->items, (_tmp0 = connection, (_tmp0 == NULL ? NULL : g_object_ref (_tmp0))));
	tuntun_connections_on_connection_added (self, connection);
	g_signal_connect (connection, "notify::status", ((GCallback) (tuntun_connections_connection_notify_property_changed)), self);
	g_signal_connect (connection, "notify::control-channel-status", ((GCallback) (tuntun_connections_connection_notify_property_changed)), self);
	g_signal_connect_object (connection, "authentication-required", ((GCallback) (_tuntun_connections_on_connection_authentication_required_tuntun_connection_authentication_required)), self, 0);
	g_signal_connect_object (connection, "authentication-failed", ((GCallback) (_tuntun_connections_on_connection_authentication_failed_tuntun_connection_authentication_failed)), self, 0);
	g_signal_connect_object (connection, "control-channel-fatal-error", ((GCallback) (_tuntun_connections_on_connection_fatal_error_tuntun_connection_control_channel_fatal_error)), self, 0);
	g_signal_connect_object (connection, "control-channel-data-received", ((GCallback) (_tuntun_connections_on_connection_data_received_tuntun_connection_control_channel_data_received)), self, 0);
}


void tuntun_connections_remove (TuntunConnections* self, TuntunConnection* connection) {
	guint _tmp0;
	guint _tmp1;
	guint _tmp2;
	guint _tmp3;
	g_return_if_fail (TUNTUN_IS_CONNECTIONS (self));
	g_return_if_fail (TUNTUN_IS_CONNECTION (connection));
	self->items = g_list_remove (self->items, connection);
	g_signal_handlers_disconnect_by_func (connection, ((void*) (tuntun_connections_connection_notify_property_changed)), self);
	g_signal_handlers_disconnect_matched (connection, G_SIGNAL_MATCH_ID | G_SIGNAL_MATCH_FUNC | G_SIGNAL_MATCH_DATA, (g_signal_parse_name ("authentication-required", TUNTUN_TYPE_CONNECTION, &_tmp0, NULL, FALSE), _tmp0), 0, NULL, ((GCallback) (_tuntun_connections_on_connection_authentication_required_tuntun_connection_authentication_required)), self);
	g_signal_handlers_disconnect_matched (connection, G_SIGNAL_MATCH_ID | G_SIGNAL_MATCH_FUNC | G_SIGNAL_MATCH_DATA, (g_signal_parse_name ("authentication-failed", TUNTUN_TYPE_CONNECTION, &_tmp1, NULL, FALSE), _tmp1), 0, NULL, ((GCallback) (_tuntun_connections_on_connection_authentication_failed_tuntun_connection_authentication_failed)), self);
	g_signal_handlers_disconnect_matched (connection, G_SIGNAL_MATCH_ID | G_SIGNAL_MATCH_FUNC | G_SIGNAL_MATCH_DATA, (g_signal_parse_name ("control-channel-fatal-error", TUNTUN_TYPE_CONNECTION, &_tmp2, NULL, FALSE), _tmp2), 0, NULL, ((GCallback) (_tuntun_connections_on_connection_fatal_error_tuntun_connection_control_channel_fatal_error)), self);
	g_signal_handlers_disconnect_matched (connection, G_SIGNAL_MATCH_ID | G_SIGNAL_MATCH_FUNC | G_SIGNAL_MATCH_DATA, (g_signal_parse_name ("control-channel-data-received", TUNTUN_TYPE_CONNECTION, &_tmp3, NULL, FALSE), _tmp3), 0, NULL, ((GCallback) (_tuntun_connections_on_connection_data_received_tuntun_connection_control_channel_data_received)), self);
	tuntun_connections_on_connection_removed (self, connection);
}


static void tuntun_connections_on_connection_data_received (TuntunConnections* self, TuntunConnection* connection, const char* data) {
	g_return_if_fail (TUNTUN_IS_CONNECTIONS (self));
	g_return_if_fail (TUNTUN_IS_CONNECTION (connection));
	g_return_if_fail (data != NULL);
	tuntun_connections_on_connection_activity (self, connection);
}


static void tuntun_connections_real_on_connection_activity (TuntunConnections* self, TuntunConnection* connection) {
	g_return_if_fail (TUNTUN_IS_CONNECTIONS (self));
	g_return_if_fail (TUNTUN_IS_CONNECTION (connection));
	g_signal_emit_by_name (G_OBJECT (self), "connection-activity", connection);
}


void tuntun_connections_on_connection_activity (TuntunConnections* self, TuntunConnection* connection) {
	TUNTUN_CONNECTIONS_GET_CLASS (self)->on_connection_activity (self, connection);
}


static void tuntun_connections_real_on_connection_fatal_error (TuntunConnections* self, TuntunConnection* connection, const char* error) {
	g_return_if_fail (TUNTUN_IS_CONNECTIONS (self));
	g_return_if_fail (TUNTUN_IS_CONNECTION (connection));
	g_return_if_fail (error != NULL);
	g_signal_emit_by_name (G_OBJECT (self), "connection-fatal-error", connection, error);
}


void tuntun_connections_on_connection_fatal_error (TuntunConnections* self, TuntunConnection* connection, const char* error) {
	TUNTUN_CONNECTIONS_GET_CLASS (self)->on_connection_fatal_error (self, connection, error);
}


static void tuntun_connections_real_on_connection_authentication_required (TuntunConnections* self, TuntunConnection* connection, TuntunAuthenticationModes mode, const char* type) {
	g_return_if_fail (TUNTUN_IS_CONNECTIONS (self));
	g_return_if_fail (TUNTUN_IS_CONNECTION (connection));
	g_return_if_fail (type != NULL);
	g_signal_emit_by_name (G_OBJECT (self), "authentication-required", connection, mode, type);
}


void tuntun_connections_on_connection_authentication_required (TuntunConnections* self, TuntunConnection* connection, TuntunAuthenticationModes mode, const char* type) {
	TUNTUN_CONNECTIONS_GET_CLASS (self)->on_connection_authentication_required (self, connection, mode, type);
}


static void tuntun_connections_real_on_connection_authentication_failed (TuntunConnections* self, TuntunConnection* connection, TuntunAuthenticationModes mode, const char* type) {
	g_return_if_fail (TUNTUN_IS_CONNECTIONS (self));
	g_return_if_fail (TUNTUN_IS_CONNECTION (connection));
	g_return_if_fail (type != NULL);
	g_signal_emit_by_name (G_OBJECT (self), "authentication-failed", connection, mode, type);
}


void tuntun_connections_on_connection_authentication_failed (TuntunConnections* self, TuntunConnection* connection, TuntunAuthenticationModes mode, const char* type) {
	TUNTUN_CONNECTIONS_GET_CLASS (self)->on_connection_authentication_failed (self, connection, mode, type);
}


static void tuntun_connections_real_on_connection_removed (TuntunConnections* self, TuntunConnection* connection) {
	g_return_if_fail (TUNTUN_IS_CONNECTIONS (self));
	g_return_if_fail (TUNTUN_IS_CONNECTION (connection));
	g_signal_emit_by_name (G_OBJECT (self), "connection-removed", connection);
}


void tuntun_connections_on_connection_removed (TuntunConnections* self, TuntunConnection* connection) {
	TUNTUN_CONNECTIONS_GET_CLASS (self)->on_connection_removed (self, connection);
}


static void tuntun_connections_real_on_connection_added (TuntunConnections* self, TuntunConnection* connection) {
	g_return_if_fail (TUNTUN_IS_CONNECTIONS (self));
	g_return_if_fail (TUNTUN_IS_CONNECTION (connection));
	g_signal_emit_by_name (G_OBJECT (self), "connection-added", connection);
}


void tuntun_connections_on_connection_added (TuntunConnections* self, TuntunConnection* connection) {
	TUNTUN_CONNECTIONS_GET_CLASS (self)->on_connection_added (self, connection);
}


static gboolean _tuntun_connections_connection_recheck_gsource_func (gpointer self) {
	return tuntun_connections_connection_recheck (self);
}


static void tuntun_connections_real_on_connection_status_changed (TuntunConnections* self, TuntunConnection* connection) {
	g_return_if_fail (TUNTUN_IS_CONNECTIONS (self));
	g_return_if_fail (TUNTUN_IS_CONNECTION (connection));
	g_signal_emit_by_name (G_OBJECT (self), "connection-status-changed", connection);
	/*see if I need to reinit the connection later*/
	if (self->priv->_timeout_id == 0 && tuntun_connection_get_control_channel_status (connection) == TUNTUN_CONNECTION_STATES_ERROR) {
		self->priv->_timeout_id = g_timeout_add_seconds (((guint) (15)), _tuntun_connections_connection_recheck_gsource_func, self);
	}
}


void tuntun_connections_on_connection_status_changed (TuntunConnections* self, TuntunConnection* connection) {
	TUNTUN_CONNECTIONS_GET_CLASS (self)->on_connection_status_changed (self, connection);
}


static gboolean tuntun_connections_connection_recheck (TuntunConnections* self) {
	gboolean result;
	g_return_val_if_fail (TUNTUN_IS_CONNECTIONS (self), FALSE);
	result = FALSE;
	{
		GList* connection_collection;
		GList* connection_it;
		connection_collection = self->items;
		for (connection_it = connection_collection; connection_it != NULL; connection_it = connection_it->next) {
			TuntunConnection* _tmp0;
			TuntunConnection* connection;
			_tmp0 = NULL;
			connection = (_tmp0 = ((TuntunConnection*) (connection_it->data)), (_tmp0 == NULL ? NULL : g_object_ref (_tmp0)));
			{
				if (tuntun_connection_get_control_channel_status (connection) == TUNTUN_CONNECTION_STATES_ERROR) {
					tuntun_connection_reinitialize (connection);
					result = TRUE;
				}
				(connection == NULL ? NULL : (connection = (g_object_unref (connection), NULL)));
			}
		}
	}
	/*ensure that
	we do this
	for 2 times since some bad concurrency could happen here*/
	if (!result) {
		self->priv->_timeout_id = ((guint) (0));
	}
	return result;
}


static void tuntun_connections_connection_notify_property_changed (TuntunConnection* connection, GParamSpec* param, TuntunConnections* connections) {
	g_return_if_fail (TUNTUN_IS_CONNECTION (connection));
	g_return_if_fail (param != NULL);
	g_return_if_fail (TUNTUN_IS_CONNECTIONS (connections));
	tuntun_connections_on_connection_status_changed (connections, connection);
}


static void tuntun_connections_xmlconfig_end_element_handler (GMarkupParseContext* context, const char* element_name, void* user_data) {
	g_return_if_fail (context != NULL);
	g_return_if_fail (element_name != NULL);
}


static void tuntun_connections_xmlconfig_text_handler (GMarkupParseContext* context, const char* text, gulong text_len, void* user_data) {
	g_return_if_fail (context != NULL);
	g_return_if_fail (text != NULL);
}


TuntunConnections* tuntun_connections_new (void) {
	TuntunConnections * self;
	self = g_object_newv (TUNTUN_TYPE_CONNECTIONS, 0, NULL);
	return self;
}


static GObject * tuntun_connections_constructor (GType type, guint n_construct_properties, GObjectConstructParam * construct_properties) {
	GObject * obj;
	TuntunConnectionsClass * klass;
	GObjectClass * parent_class;
	TuntunConnections * self;
	klass = TUNTUN_CONNECTIONS_CLASS (g_type_class_peek (TUNTUN_TYPE_CONNECTIONS));
	parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
	obj = parent_class->constructor (type, n_construct_properties, construct_properties);
	self = TUNTUN_CONNECTIONS (obj);
	{
		GList* _tmp0;
		_tmp0 = NULL;
		self->items = (_tmp0 = NULL, (self->items == NULL ? NULL : (self->items = (_g_list_free_g_object_unref (self->items), NULL))), _tmp0);
		tuntun_connections_load (self);
	}
	return obj;
}


static void tuntun_connections_class_init (TuntunConnectionsClass * klass) {
	tuntun_connections_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (TuntunConnectionsPrivate));
	G_OBJECT_CLASS (klass)->constructor = tuntun_connections_constructor;
	G_OBJECT_CLASS (klass)->dispose = tuntun_connections_dispose;
	TUNTUN_CONNECTIONS_CLASS (klass)->on_connection_activity = tuntun_connections_real_on_connection_activity;
	TUNTUN_CONNECTIONS_CLASS (klass)->on_connection_fatal_error = tuntun_connections_real_on_connection_fatal_error;
	TUNTUN_CONNECTIONS_CLASS (klass)->on_connection_authentication_required = tuntun_connections_real_on_connection_authentication_required;
	TUNTUN_CONNECTIONS_CLASS (klass)->on_connection_authentication_failed = tuntun_connections_real_on_connection_authentication_failed;
	TUNTUN_CONNECTIONS_CLASS (klass)->on_connection_removed = tuntun_connections_real_on_connection_removed;
	TUNTUN_CONNECTIONS_CLASS (klass)->on_connection_added = tuntun_connections_real_on_connection_added;
	TUNTUN_CONNECTIONS_CLASS (klass)->on_connection_status_changed = tuntun_connections_real_on_connection_status_changed;
	g_signal_new ("connection_added", TUNTUN_TYPE_CONNECTIONS, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__OBJECT, G_TYPE_NONE, 1, TUNTUN_TYPE_CONNECTION);
	g_signal_new ("connection_removed", TUNTUN_TYPE_CONNECTIONS, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__OBJECT, G_TYPE_NONE, 1, TUNTUN_TYPE_CONNECTION);
	g_signal_new ("connection_status_changed", TUNTUN_TYPE_CONNECTIONS, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__OBJECT, G_TYPE_NONE, 1, TUNTUN_TYPE_CONNECTION);
	g_signal_new ("connection_activity", TUNTUN_TYPE_CONNECTIONS, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__OBJECT, G_TYPE_NONE, 1, TUNTUN_TYPE_CONNECTION);
	g_signal_new ("connection_fatal_error", TUNTUN_TYPE_CONNECTIONS, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_user_marshal_VOID__OBJECT_STRING, G_TYPE_NONE, 2, TUNTUN_TYPE_CONNECTION, G_TYPE_STRING);
	g_signal_new ("authentication_required", TUNTUN_TYPE_CONNECTIONS, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_user_marshal_VOID__OBJECT_ENUM_STRING, G_TYPE_NONE, 3, TUNTUN_TYPE_CONNECTION, TUNTUN_TYPE_AUTHENTICATION_MODES, G_TYPE_STRING);
	g_signal_new ("authentication_failed", TUNTUN_TYPE_CONNECTIONS, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_user_marshal_VOID__OBJECT_ENUM_STRING, G_TYPE_NONE, 3, TUNTUN_TYPE_CONNECTION, TUNTUN_TYPE_AUTHENTICATION_MODES, G_TYPE_STRING);
}


static void tuntun_connections_instance_init (TuntunConnections * self) {
	self->priv = TUNTUN_CONNECTIONS_GET_PRIVATE (self);
	self->priv->_timeout_id = ((guint) (0));
}


static void tuntun_connections_dispose (GObject * obj) {
	TuntunConnections * self;
	self = TUNTUN_CONNECTIONS (obj);
	self->priv->_config_file = (g_free (self->priv->_config_file), NULL);
	(self->items == NULL ? NULL : (self->items = (_g_list_free_g_object_unref (self->items), NULL)));
	G_OBJECT_CLASS (tuntun_connections_parent_class)->dispose (obj);
}


GType tuntun_connections_get_type (void) {
	static volatile gsize tuntun_connections_type_id = 0;
	if (g_once_init_enter (&tuntun_connections_type_id)) {
		GType tuntun_connections_type_id_temp;
		static const GTypeInfo g_define_type_info = { sizeof (TuntunConnectionsClass), (GBaseInitFunc) NULL, (GBaseFinalizeFunc) NULL, (GClassInitFunc) tuntun_connections_class_init, (GClassFinalizeFunc) NULL, NULL, sizeof (TuntunConnections), 0, (GInstanceInitFunc) tuntun_connections_instance_init };
		tuntun_connections_type_id_temp = g_type_register_static (G_TYPE_OBJECT, "TuntunConnections", &g_define_type_info, 0);
		g_once_init_leave (&tuntun_connections_type_id, tuntun_connections_type_id_temp);
	}
	return tuntun_connections_type_id;
}


static int _vala_strcmp0 (const char * str1, const char * str2) {
	if (str1 == NULL) {
		return -(str1 != str2);
	}
	if (str2 == NULL) {
		return (str1 != str2);
	}
	return strcmp (str1, str2);
}



static void g_cclosure_user_marshal_VOID__OBJECT_STRING (GClosure * closure, GValue * return_value, guint n_param_values, const GValue * param_values, gpointer invocation_hint, gpointer marshal_data) {
	typedef void (*GMarshalFunc_VOID__OBJECT_STRING) (gpointer data1, gpointer arg_1, const char* arg_2, gpointer data2);
	register GMarshalFunc_VOID__OBJECT_STRING callback;
	register GCClosure * cc;
	register gpointer data1, data2;
	cc = ((GCClosure *) (closure));
	g_return_if_fail (n_param_values == 3);
	if (G_CCLOSURE_SWAP_DATA (closure)) {
		data1 = closure->data;
		data2 = param_values->data[0].v_pointer;
	} else {
		data1 = param_values->data[0].v_pointer;
		data2 = closure->data;
	}
	callback = ((GMarshalFunc_VOID__OBJECT_STRING) ((marshal_data ? marshal_data : cc->callback)));
	callback (data1, g_value_get_object (param_values + 1), g_value_get_string (param_values + 2), data2);
}


static void g_cclosure_user_marshal_VOID__OBJECT_ENUM_STRING (GClosure * closure, GValue * return_value, guint n_param_values, const GValue * param_values, gpointer invocation_hint, gpointer marshal_data) {
	typedef void (*GMarshalFunc_VOID__OBJECT_ENUM_STRING) (gpointer data1, gpointer arg_1, gint arg_2, const char* arg_3, gpointer data2);
	register GMarshalFunc_VOID__OBJECT_ENUM_STRING callback;
	register GCClosure * cc;
	register gpointer data1, data2;
	cc = ((GCClosure *) (closure));
	g_return_if_fail (n_param_values == 4);
	if (G_CCLOSURE_SWAP_DATA (closure)) {
		data1 = closure->data;
		data2 = param_values->data[0].v_pointer;
	} else {
		data1 = param_values->data[0].v_pointer;
		data2 = closure->data;
	}
	callback = ((GMarshalFunc_VOID__OBJECT_ENUM_STRING) ((marshal_data ? marshal_data : cc->callback)));
	callback (data1, g_value_get_object (param_values + 1), g_value_get_enum (param_values + 2), g_value_get_string (param_values + 3), data2);
}



